#include <iostream>
#include <stack>
#include <string>
#include <stdexcept>

#include <string.h>
#include <stdio.h>

#include "format.hpp"

using namespace fortran_format;

void delete_ast(ast_node* n) {
 if (!n) return;
 if (n->letter == '(') delete_ast(n->group);
 ast_node* next = n->next;
 delete n;
 delete_ast(next);
};

inline bool is_digit(char c) {
 return c >= '0' && c <= '9';
};

int read_integer(const char** string) {
 int i = 0;
 const char* s = *string;
 while (*s >= '0' && *s <= '9')
  i = i*10 + *s++ - '0';
 *string = s;
 return i;
};

ast_node* do_parse_format(const char** format, int recursion_level) {
 ast_node* head = NULL;
 const char* f = *format;
 try {
  ast_node* tail = NULL;

  while (*f) {
   int count = is_digit(*f) ? read_integer(&f) : 1;
   char letter = *f;
   int width = 0;
   ast_node* group = NULL;

   switch (*f++) {
    case 'I':
    case 'A':
     if (!is_digit(*f)) goto error;
     width = read_integer(&f);
     break;
    case 'E':
    case 'F':
     if (!is_digit(*f)) goto error;
     width = read_integer(&f);
     if (*f++ != '.' || !is_digit(*f)) goto error;
     while (is_digit(*++f));
     break;
    case 'X':
     break;
    case '(':
     group = do_parse_format(&f, recursion_level+1);
     break;
    default:
     goto error;
   };

   if (!head) 
    head = tail = new ast_node;
   else {
    tail->next = new ast_node;
    tail = tail->next;
   };
   tail->next = NULL;
   tail->count = count;
   tail->letter = letter;
   if (letter == '(') 
    tail->group = group;
   else if (width == 0 && letter != 'X')
    goto error;
   else
    tail->width = width;

   if (*f == ',') 
    ++f;
   else if (*f == ')' && recursion_level != 0) {
    *format = f+1;
    return head;
   } else if (*f)
    goto error;
  };

  if (recursion_level != 0) goto error;
  return head;

error:
  throw *format;

 } catch(const char* f) {
  if (head) delete_ast(head);
  if (recursion_level == 0)
   throw formatted_exception("Invalid Fortran format \"%1%\"") << *format;
  throw;
 };
};

ast_node* parse_format(const char* format) {
 return do_parse_format(&format, 0);
};

ast& ast::parse(const char* format) {
 if (own && head) delete_ast(head);
 head = parse_format(format);
 own = true;
 return *this;
};

ast::ast(): head(NULL), own(false) {};

ast::ast(const char* format): head(parse_format(format)), own(true) {};

ast::ast(ast& t): head(t.head), own(false) {};

ast::~ast() {
 if (own) delete_ast(head);
};

ast& ast::operator=(ast& t) {
 if (own) delete_ast(head);
 head = t.head;
 own = false;
 return *this;
};

ast_node& ast::operator*() {
 return *head;
};

ast_node* ast::operator->() {
 return head;
};

ast::operator ast_node*() {
 return head;
};

void reader::init() {
 state.esp = t;
 state.count = t->count;
 newline = false;
};

reader::reader(): s(NULL), newline(false) {
 state.esp = NULL;
};

reader::reader(const char* format)
 : t(format), s(NULL) {
  init();
};

reader::reader(const char* format, std::streambuf& s)
 : t(format), s(&s) {
  init();
};

reader::reader(ast& t): t(t), s(NULL) {
 init();
};

reader::reader(ast& t, std::streambuf& s): t(t), s(&s) {
 init();
};

void reader::advance_state() {
 if (!state.esp) {
  state.esp = t;
  state.count = t->count;
 };

 if (state.count == 0) {
  if (state.esp->next) state.esp = state.esp->next;
  else if (stack.empty()) {
   if (!newline) skip_rest_of_line();
   else newline = false;
   state.esp = t;
  } else {
   state = stack.top();
   stack.pop();
   advance_state();
   return;
  };
  state.count = state.esp->count;
 };

 if (state.esp->letter == 'X') {
  if (!newline) {
   int c;
   do c = s->sbumpc();
   while (--state.count && c != EOF && c != '\n');
   if (c == '\r') {
    newline = s->sbumpc() == '\n';
    if (!newline) s->sungetc();
   } else newline = c == '\n';
  };
  state.count = 0;
  advance_state();
  return;
 };

 --state.count;

 if (state.esp->letter == '(') {
  stack.push(state);
  state.esp = state.esp->group;
  state.count = state.esp->count;
  advance_state();
 };
};

reader& reader::reset() {
 state.esp = t;
 state.count = t->count;
 newline = false;
 return *this;
};

reader& reader::end() {
 if (!newline) skip_rest_of_line();
 else newline = false;
 return *this;
};

reader& reader::next_line() {
 skip_rest_of_line();
 reset();
 return *this;
};

bool is_space(const char* s) {
 while (*s) if (*s != ' ') return false;
 return true;
};

int reader::next_char() {
 int result = s->sbumpc();

 if (result == '\r') {
  result = s->sbumpc();
  if (result != '\n') {
   s->sungetc();
   return '\r';
  };
 };

 if (result == '\n') {
  newline = true;
  return EOF;
 };

 return result;
};

void reader::skip_rest_of_line() {
 int c;
#ifdef DEBUG
 debug << "Fortran format reader: skipping line: ";
#endif
 do {
  c = s->sbumpc();
#ifdef DEBUG
  debug << c;
#endif
 } while (c != EOF && c != '\n');
};

reader& reader::operator>>(int& x) {
 advance_state();

 if (state.esp->letter != 'I')
  throw error(state.esp->letter, "int");

 if (newline) {
  x = 0;
  return *this;
 };
 
 char buf[state.esp->width+1];
 char* b = buf;
 for (int i = 0; i < state.esp->width; ++i) {
  int c = next_char();
  if (c == EOF) break;
  if (c != ' ') *b++ = c;
 };
 *b = 0;

 if (*buf) {
  char* e;
  x = strtol(buf, &e, 10);
  if (e < b)
   throw formatted_exception("Bad value during integer[%1%] read: `%2%'")
    << state.esp->width << (char*)buf;
 } else
  x = 0;

#ifdef DEBUG
 debug << "Fortran format reader: " 
  << state.esp->letter << state.esp->width << ": `" 
  << &buf[0] << "' -> " << x << '\n';
#endif
 return *this;
};

reader& reader::operator>>(double& x) {
 advance_state();

 if (state.esp->letter != 'E' && state.esp->letter != 'F')
  throw error(state.esp->letter, "double");

 if (newline) {
  x = 0;
  return *this;
 };

 char buf[state.esp->width+1];
 char* b = buf;
 for (int i = 0; i < state.esp->width; ++i) {
  int c = next_char();
  if (c == EOF) break;
  if (c != ' ') *b++ = c;
 };
 *b = 0;

 if (*buf) {
  char* e;
  x = strtod(buf, &e);
  if (e < b)
   throw formatted_exception("Bad value during double[%1%] read: `%2%'")
    << state.esp->width << (char*)buf;
 } else
  x = 0;

#ifdef DEBUG
 debug << "Fortran format reader: " 
  << state.esp->letter << state.esp->width << ": `" 
  << &buf[0] << "' -> " << x << '\n';
#endif
 return *this;
};

reader& reader::operator>>(std::string& x) {
 advance_state();

 if (state.esp->letter != 'A')
  throw error(state.esp->letter, "string");

 x.resize(state.esp->width);
 
 if (newline) {
  x.replace(x.begin(), x.end(), x.size(), ' ');
  return *this;
 };

 int i = 0;
 for (; i < state.esp->width; ++i) {
  int c = next_char();
  if (c == EOF) break;
  x[i] = c;
 };
 int n = state.esp->width - i;
 x.replace(i, n, n, ' ');

#ifdef DEBUG
 debug << "Fortran format reader: " 
  << state.esp->letter << state.esp->width << ": `" 
  << x << "'\n";
#endif
 return *this;
};

error::error(char letter, const char* got):
 formatted_exception(
   "Format error: expected %1% in formatted transfer, got %2%"
 ) {
 switch (letter) {
  case 'I':
   *this << "int";
   break;
  case 'E':
  case 'F':
   *this << "double";
   break;
  case 'A':
   *this << "string";
   break;
 };
 *this << got;
};
