#ifndef FORTRAN_FORMAT_HEADER
#define FORTRAN_FORMAT_HEADER

#include <iostream>
#include <stack>
#include <string>

#include <stdio.h>

#include "formatted_exception.hpp"

#ifdef DEBUG
#include "debug.hpp"
#endif

namespace fortran_format {

struct ast_node;

struct ast_node {
 char letter;
 int count;
 union {
  int width;
  ast_node* group;
 };
 ast_node* next;
};

class ast {
 public:
  ast_node* head;
  bool own;

  ast();
  ast(const char* format);
  ast(ast& t);
  
  ~ast();

  ast& parse(const char* format);

  ast& operator=(ast& t);
  ast_node& operator*();
  ast_node* operator->();
  operator ast_node*();
};

class reader {
 public:
  struct state_t {
   int count;
   ast_node* esp;
  };

  ast t;
  std::streambuf* s;
  std::stack<state_t> stack;
  state_t state;
  bool newline;
  size_t max_values_per_line;
  size_t values_read;

  void init();

  reader();
  reader(const char* format);
  reader(const char* format, std::streambuf& s);
  reader(ast& t);
  reader(ast& t, std::streambuf& s);

  int next_char();
  void skip_rest_of_line();
  void scan_ast();

  void advance_state();

  reader& reset();
  reader& next_line();
  reader& set_ast();

  inline reader& begin() { return reset(); };
  reader& end();

  reader& operator>>(int& x);
  reader& operator>>(double& x);
  reader& operator>>(std::string& x);
  template <int n> reader& operator>>(char (&x)[n]);
};

class error: public formatted_exception {
 public:
  error(char letter, const char* got);
};

template <int n> reader& reader::operator>>(char (&x)[n]) {
 advance_state();

 if (state.esp->letter != 'A')
  throw error(state.esp->letter, "char[]");

 if (newline) {
  memset(x, ' ', n-1);
  x[n-1] = 0;
  return *this;
 };

 char* buf = n > (size_t)state.esp->width ? x : new char[state.esp->width];

 int i = 0;
 while (i < state.esp->width) {
  int c = next_char();
  if (c == EOF) break;
  buf[i++] = c;
 };

 if ((size_t)i < n && i <= state.esp->width) {
  if (buf != x) memcpy(x, buf, i);
  memset(x+i, ' ', n-i);
 } else
  memcpy(x, buf + i - sizeof(i) + 2, i - sizeof(i) + 1);
 x[n-1] = 0;

 if (buf != x) delete[] buf;

#ifdef DEBUG
 debug << "Fortran format reader: " 
  << state.esp->letter << state.esp->width << ": `" 
  << &buf[0] << "' -> " << x << '\n';
#endif
 return *this;
};

};

#endif
