#include <vector>
#include <iostream>
#include <iomanip>
#include <fstream>

#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/lexical_cast.hpp>

#include <math.h>

#include <gsl/gsl_errno.h>

#include "formatted_exception.hpp"
#include "subgroups.hpp"
#include "program_options.hpp"
#include "abbn.hpp"

#ifdef DEBUG
#include "debug.hpp"
#endif

// group_data holds group input data
struct group_data {
 int group;
 double total;
 std::vector<double> partials;

 // When working with ABBN files, the following fields hold average cross
 // sections at 300 K.  The result stored in the output are subgroup parameters
 // divided by these.
 double total0;
 std::vector<double> partials0;

 int ignore;
 std::vector<subgroups::knot> knots;
 size_t K;
};

// Convert blocked cross-sections into momenta
void convert_bxs_to_knots(
  std::vector<subgroups::knot>& knots, double nonresonant
) {
 for (std::vector<subgroups::knot>::iterator k = knots.begin();
   k != knots.end(); ++k) {
  k->total1 = 1. / (k->dilution + k->total1);
  k->total2 = k->total1 / (k->dilution + k->total2);
  k->dilution += nonresonant;
  for (std::vector<double>::iterator x = k->partial.begin();
    x != k->partial.end(); ++x)
   *x *= k->total1;
 };
};

group_data& convert_bxs_to_knots(group_data& gd) {
 double nonresonant = gd.total;
 for (std::vector<double>::const_iterator x = gd.partials.begin();
   x != gd.partials.end(); ++x)
  nonresonant -= *x;
 if (nonresonant < 0) nonresonant = 0;
 convert_bxs_to_knots(gd.knots, nonresonant);
 return gd;
};

#ifndef WIN32
#define win_fix_exponent(x) (x)
#else
// Workaround for Windows retarted C library that insists on printing three
// digits in exponents even if it results in leading zeroes, like this:
// 1.000e+001. See 
// http://groups.google.com/group/comp.lang.c++/browse_thread/thread/624b679a4faf03d 
// for details

std::string win_fix_exponent(double x) {
 std::stringstream s;
 s << std::setprecision(5) << std::scientific << x;
 std::string result;
 s >> result;
 int i = result.size()-3;
 if (i > 2
     && result[i] == '0'
     && ((result[i-1] == '+' || result[i-1] == '-')
         && (result[i-2] == 'e' || result[i-2] == 'E')
         || (result[i-1] == 'e' || result[i-1] == 'E')))
  result.erase(i, 1);
 return result;
};
#endif

void print_abbn_header(
  size_t size, 
  char nam[10], 
  char bib[5], 
  double awr, 
  double temperature, 
  int lv, 
  bool total
) {
 std::cout 
  << "NAM=" << std::left << std::setw(9) << nam
  << "LIB=" << std::left << std::setw(4) << bib
  << "  MF=306  MT=  0  AWR=" << std::scientific << std::setw(12) 
     << win_fix_exponent(awr)
  << "  TEM=" << std::scientific << std::setw(13) 
     << win_fix_exponent(temperature) 
  << std::endl
  << "LV =" << std::right << std::setw(3) << lv 
  << "      LT =" << std::right << std::setw(4) << size 
  << "  LC=  " << (total ? '7' : '6')
  << "  LS=  " << (total ? '7' : '6')
  << "  LF = (" << (lv == 20 ? "2I2" : "I4") << ',' << (total ? '5' : '4') << "E12.0)" 
     << std::endl
  << "* Gr probability     capture     elastic     fission" 
  << (total ? "       total" : "") << std::endl;
};

void print_abbn_subgroups(
  size_t group, 
  const group_data& gd,
  const std::vector<subgroup>& subgroups, 
  bool fissile,
  bool total,
  const std::vector<int>& abbn_structure
) {
 size_t g = group;
 int wg;
 if (!abbn_structure.empty()) {
  wg = abbn_structure[group-1];
  while (g-- > 0 && abbn_structure[g] == wg);
  g = group-1 - g;
 };

 for (std::vector<subgroup>::const_iterator sg = subgroups.begin();
   sg != subgroups.end(); ++sg) {
  if (sg == subgroups.begin())
   if (abbn_structure.empty())
    std::cout << std::setw(4) << g;
   else
    std::cout << std::setw(2) << wg << std::setw(2) << g;
  else
   std::cout << "    ";

  std::cout
   << std::setw(12) << std::setprecision(6) << std::fixed 
   << sg->probability << std::setprecision(5) << std::scientific;
  for (size_t i = 0; i < (fissile ? 3 : 2); ++i)
   std::cout 
    << std::setw(12) << win_fix_exponent(sg->partial[i] / gd.partials0[i]);
  if (!fissile) std::cout << std::setw(12) << win_fix_exponent(1.);
  if (total) std::cout << std::setw(12) << win_fix_exponent(sg->total / gd.total0);
  std::cout << std::endl;
 };
};

// Check if inelastic (unspecified) cross section varies too much
group_data& test_inelastic(group_data& gd, double maxerror) {
 double inelastic = gd.total;
 double min = gd.total; // the smallest average partial cross section
 for (std::vector<double>::const_iterator p = gd.partials.begin(); 
   p != gd.partials.end(); ++p) {
  inelastic -= *p;
  if (*p < min) min = *p;
 };
 if (inelastic < maxerror * min) 
  // The inelastic cross section is negligible
  return gd;

 // Test the self-shielded inelastic cross section dependence of dilution
 // cross section
 std::vector<double> blocked_inelastic(gd.knots.size());
 bool negligible = true;
 for (size_t i = 0; i < gd.knots.size(); ++i) {
  // Compute the self-shielded inelastic cross section as a difference between
  // total and sum of partials
  double bi = gd.knots[i].total1;
  for (std::vector<double>::const_iterator p = gd.knots[i].partial.begin();
    p != gd.knots[i].partial.end(); ++p)
   bi -= *p;

  blocked_inelastic[i] = bi;
  negligible = negligible && fabs(bi / inelastic - 1) < maxerror;
 };

 if (!negligible) {
  // Take the inelastic cross section into account
  gd.partials.push_back(blocked_inelastic.back());
  for (size_t i = 0; i < gd.knots.size(); ++i)
   gd.knots[i].partial.push_back(blocked_inelastic[i]);
  gd.ignore = gd.partials.size()-1;
 };

 return gd;
};

// Prints an array
template <typename T> std::istream& operator>>(
  std::istream& s, std::vector<T>& v
) {
 v.clear();
 while (true) {
  T x;
  s >> x;
  if (!s) break;
  v.push_back(x);
 };
 return s;
};

void usage(const boost::program_options::options_description& desc) {
 std::cout
  << "This program calculates subgroups parameters for neutron cross-sections\n"
     "Usage: subgroups [parameters] abbn-file \n"
  << desc << '\n';
 exit(0);
};

// Parse command line arguments
boost::program_options::variables_map parse_arguments(int argc, char** argv) {
 using namespace boost::program_options;
 // Command line arguments definition
 my_option_description desc[] = {
  { "help", 'h', NULL, NULL, "Print help message" },
#ifdef DEBUG
  { "debug", 'D', NULL, NULL, "Turn on verbose debugging output" },
#endif
  { "dilutions", 'd', "SUBGROUPS_DILUTIONS", value<std::string>(),
   "File with dilution cross-sections" },
  { "fissile", 'f', NULL, NULL, "Take the fission cross-section into account" },
  { "total", 'T', NULL, NULL, "Add total cross-section to output" },
  { "error", 'e', "SUBGROUPS_ERROR", value<double>()->default_value(1E-2),
   "Maximum permissible error in blocked cross-section" },
  { "temperature", 't', NULL, value<double>()->default_value(300),
   "Seek that temperature in the input files" },
  { "structure", 's', NULL, value<std::string>(),
   "ABBN file to read subgroup structure from" }
 };
 my_positional_option_description pos 
  = { "data", NULL, value<std::vector<std::string> >()->composing(), 2 };

 return parse_args(argc, argv, desc, sizeof(desc), &pos, usage);
};

inline double& knot_xsection(subgroups::knot& k, int index) {
 switch (index) {
  case 0: return k.total1;
  case 1: return k.total2;
  default: return k.partial[index-2];
 };
};

inline double& group_xsection(group_data& g, int index) {
 if (index == 0) return g.total;
 else return g.partials[index-2];
};

// Reads mean cross sections from ABBN file
void read_abbn_xsections(
  iabbn& abbn, 
  std::vector<int>& abbn_structure, 
  std::vector<group_data>& result, 
  std::vector<double>& inelastic,
  bool fissile
) {
 using namespace boost::lambda;

#ifdef DEBUG
 debug << "Loading table MF=301, MT=0\n";
#endif
 if (!abbn.find(&_1->*&abbn::header::mf == 301 && &_1->*&abbn::header::mt == 0))
  throw formatted_exception(
    "Could not find section MF = 301, MT = 0 in ABBN file"
  );

 result.resize(abbn.h.lt);
 inelastic.resize(abbn.h.lt);

 /* ABBN format is loosely specified in terms of group numbers. When LV = 20,
  * two numbers are used to denote a group, and the second number sometimes
  * starts with 1 for every change of the first number, and sometimes it
  * increments all they way to the end. In order to deal with it, first numbers
  * are saved in abbn_structure array so we can use it to print subgroups data
  * later.
  */

 if (abbn.h.lv == 20)
  abbn_structure.resize(abbn.h.lt);

 for (size_t g = 0; g < (size_t)abbn.h.lt; ++g) {
  group_data& r = result[g];
  if (!abbn_structure.empty())
   abbn >> abbn_structure[g];
  abbn >> r.group >> r.total;
  r.total0 = r.total;
  r.partials.resize(fissile ? 3 : 2);
  abbn >> r.partials[0];              // capture
  if (fissile) abbn >> r.partials[2]; // fission
  else {
   double fission;
   abbn >> fission;
  };
  abbn >> r.partials[1];              // elastic
  abbn >> inelastic[g];               // inelastic
  r.partials0 = r.partials;

  // No need in the rest of the data
  abbn.next_line();
  abbn.reader.reset();

  r.ignore = -1;
 };
#ifdef DEBUG
 debug << "Table MF=301, MT=0 is loaded\n";
#endif
};

void read_abbn_knots(
  iabbn& abbn, 
  std::vector<group_data>& result, 
  double temperature
) {
 using namespace boost::lambda;

 int mts[] = { -1, 1, 102, 2, 18 }; // cross sections ABBN MT indexes
 for (size_t i = 0; i < 2 + result[0].partials.size(); ++i) {
#ifdef DEBUG
  debug << "Loading table MF=304, MT=" << mts[i] << ", TEM=" 
   << temperature << '\n';
#endif
  if (!abbn.find(
       &_1->*&abbn::header::mf == 304 
       && &_1->*&abbn::header::mt == mts[i] 
       && &_1->*&abbn::header::tem == temperature
     ))
   throw formatted_exception(
     "Could not find section MF = 304, MT = %1%, TEM = %2% in ABBN file"
   ) << mts[i] << temperature;

  for (int j = 0; j < abbn.h.lt; ++j) {
   int group;
   if (abbn.h.lv == 20) abbn >> group;
   abbn >> group;
   if (group < 1)
    throw formatted_exception(
      "Malformed self-shielding factors in ABBN file: in section MF = 304, "
      "MT = %1%, TEM = %2% there is a non-positive group: %3%"
    ) << mts[i] << temperature << group;
   if ((size_t)group > result.size())
    throw formatted_exception(
      "Inconsistent input data: in total there are %1% groups in section "
      "MF = 301, MT = 0, but section MF = 304, MT = %2%, TEM = %3% has "
      "group %4%"
    ) << result.size() << mts[i] << temperature << group;
   --group;

   double ssf;
   for (std::vector<subgroups::knot>::iterator k = result[group].knots.begin();
     k != result[group].knots.end(); ++k) {
    abbn >> ssf;
    knot_xsection(*k, i) *= ssf;
   };

   // Apply self-shielding factors for infinite dilution to take into account
   // Doppler correction
   // XXX: assumes that the last knot holds the maximum dilution cross section
   if (i != 1) group_xsection(result[group], i) *= ssf;

   abbn.next_line();
   abbn.reader.reset();
  };
#ifdef DEBUG
  debug << "Table MF=304, MT=" << mts[i] << ", TEM=" << temperature 
   << " is loaded\n";
#endif
 };
};

void read_abbn_subgroup_structure(
  iabbn& abbn, std::vector<group_data>& result
) {
 using namespace boost::lambda;

#ifdef DEBUG
 debug << "Loading table MF = 306, MT = 0\n";
#endif
 if (!abbn.find(&_1->*&abbn::header::mf == 306 && &_1->*&abbn::header::mt == 0))
  throw formatted_exception("Could not find section MF = 306, MT = 0 in ABBN file");

 int group = -1;
 for (int i = 0; i < abbn.h.lt; ++i) {
  if (abbn.h.lv == 20) {
   int w, n;
   abbn >> w >> n;
   if (w > 0 && n > 0) group = abbn::group(w, n)-1;
  } else {
   int g;
   abbn >> g;
   if (g > 0) group = g-1;
  };

  abbn.next_line();
  abbn.reader.reset();
  ++result.at(group).K;
 };
#ifdef DEBUG
 debug << "Table MF = 306, Mt = 0 is loaded\n";
#endif
};

// Reads data in ABBN format
std::vector<group_data> read_abbn_input(
  boost::program_options::variables_map options,
  char* nam,
  char* bib,
  double& awr,
  std::vector<int>& abbn_structure
) {

 boost::program_options::variables_map::const_iterator option 
  = options.find("error");
 if (option == options.end())
  throw std::runtime_error(
    "SUBGROUPS_ERROR environment is not set. Please define maximum permissible "
    "error in blocked cross-sections"
  );
 double maxerror = option->second.as<double>();

 std::string abbn_filename 
  = options["data"].as<std::vector<std::string> >().front();

 bool fissile = options.find("fissile") != options.end();
 double temperature = options["temperature"].as<double>();

 option = options.find("dilutions");
 if (option == options.end())
  throw formatted_exception("Please specify dilutions file");
 std::vector<double> dilutions;
 {
  std::ifstream f(option->second.as<std::string>().c_str());
  if (!f)
   throw formatted_exception("Cannot open dilutions file %1%")
    << option->second.as<std::string>();
  f >> dilutions;
 }

 iabbn abbn(abbn_filename.c_str());
 std::vector<group_data> result;
 std::vector<double> inelastic;

 read_abbn_xsections(abbn, abbn_structure, result, inelastic, fissile);

 // Store NAM, BIB and AWR ABBN header entries to use in the output
 strncpy(nam, abbn.h.nam, sizeof(abbn.h.nam));
 strncpy(bib, abbn.h.bib, sizeof(abbn.h.bib));
 awr = abbn.h.awr;

 // Allocate memory for knots
 for (std::vector<group_data>::iterator r = result.begin(); 
   r != result.end(); ++r) {
  r->knots.resize(dilutions.size());
  for (size_t i = 0; i < dilutions.size(); ++i) {
   subgroups::knot& k = r->knots[i];
   k.dilution = dilutions[i];
   k.total1 = k.total2 = r->total;
   k.partial = r->partials;
  };
 };

 read_abbn_knots(abbn, result, temperature);

 for (size_t g = 0; g < result.size(); ++g) {
  if (inelastic[g] != 0) 
   test_inelastic(result[g], maxerror);

  convert_bxs_to_knots(result[g]);
 };

 for (std::vector<group_data>::iterator r = result.begin(); 
   r != result.end(); ++r)
  r->K = 0;
 option = options.find("structure");
 if (option != options.end()) {
  iabbn s(option->second.as<std::string>().c_str());
  read_abbn_subgroup_structure(s, result);
 };

 return result;
};

// Abstract interface class for subgroups-related input/output
class iosubgroups {
 public:
  virtual ~iosubgroups() {};
  // Reads group
  virtual group_data get() = 0;
  // Prints group
  virtual void set(
    const group_data& g, const std::vector<subgroup>& subgroups
  ) = 0;
  // Are all groups read yet?
  virtual bool done() const = 0;
};

// ABBN format interface
class iosubgroups_abbn: public iosubgroups{
 public:
  char nam[10]; // isotope name
  char bib[5];  // library
  double awr;   // isotope mass (in neutron masses)
  std::vector<int> structure;     // ABBN wide group structure
  std::vector<group_data> groups; // the data
  size_t group;                   // current group
  std::vector<std::vector<subgroup> > result;
  double temperature;
  bool fissile;
  bool total;

  // Reads the input file
  iosubgroups_abbn(boost::program_options::variables_map& options) : group(0) {
   groups = read_abbn_input(options, nam, bib, awr, structure);
   result.resize(groups.size());
   temperature = options["temperature"].as<double>();
   fissile = options.find("fissile") != options.end();
   total   = options.find("total") != options.end();
  };

  // Prints results
  ~iosubgroups_abbn() {

   // Cut groups with no subgroups in the beginning and in the end
   size_t start = 0;
   while (start < result.size() && result[start].size() == 1) ++start;
   size_t end = result.size()-1;
   while (end > start && result[end].size() == 1) --end;

   // If wide groups are used, adjust cutting to wide group boundaries
   if (!structure.empty()) {
    int wg = structure[start];
    while (start > 0 && structure[start-1] == wg) --start;
    wg = structure[end];
    while (end < result.size()-1 && structure[end+1] == wg) ++end;
   };

   // Print ABBN header
   size_t size = 0;
   for (size_t g = start; g <= end; ++g)
    size += result[g].size();
   print_abbn_header(
     size, nam, bib, awr, temperature, structure.empty() ? 10 : 20, total
   );

   // Print data
   for (size_t g = start; g <= end; ++g)
    print_abbn_subgroups(g+1, groups[g], result[g], fissile, total, structure);
  };

  group_data get() {
   return groups[group++];
  };

  void set(const group_data& gd, const std::vector<subgroup>& subgroups) {
   result[gd.group-1] = subgroups;
  };

  bool done() const {
   return group >= groups.size();
  };
};

// Free format interface
class iosubgroups_text: public iosubgroups {
 public:
  std::istream* input;
  std::ostream* output;
  size_t group;
  double maxerror;

  // Initialize input and output with stdin and stdout
  iosubgroups_text(boost::program_options::variables_map& options) {
   input  = &std::cin;
   output = &std::cout;
   *input >> group;
   maxerror = options["error"].as<double>();
  };

  // Reads an array of doubles from input
  std::vector<double> read_vector() {
   std::stringstream ss;
   std::string s;
   std::getline(*input, s);
   ss << s;
   std::vector<double> result;
   ss >> result;
   return result;
  };

  group_data get() {
   group_data result;
   result.group = group;
   *input >> result.K >> result.total;
   result.partials = read_vector();

   // Read group data until group number changes
   while (*input >> group && group == result.group) {
    subgroups::knot k;
    *input >> k.dilution >> k.total1 >> k.total2;
    k.partial = read_vector();

    // Convert self-shielding factors to self-shielded cross sections
    k.total1 *= result.total;
    k.total2 *= result.total;
    if (k.partial.size() != result.partials.size())
     throw formatted_exception("Inconsistent input data for group %1%") 
           << group;
    for (size_t x = 0; x < result.partials.size(); ++x)
     k.partial[x] *= result.partials[x];

    result.knots.push_back(k);
   };
   result.ignore = -1;
   test_inelastic(result, maxerror);
   return convert_bxs_to_knots(result);
  };

  void set(const group_data& gd, const std::vector<subgroup>& subgroups) {
   for (size_t k = 0; k < subgroups.size(); ++k) {
    if (k == 0)
     *output << "- " << std::setw(3) << gd.group;
    else
     *output << "     ";
    *output << std::setw(3) << k+1
      << std::setw(12) << subgroups[k].probability
      << std::setw(12) << subgroups[k].total;
    for (size_t x = 0; x < subgroups[k].partial.size(); ++x)
     if ((int)x != gd.ignore) *output << std::setw(12) << subgroups[k].partial[x];
    *output << std::endl;
   };
  };

  bool done() const {
   return !*input;
  };
};


// Error handler for GNU Scientific Library. See GSL documentation
void _gsl_error_handler(
  const char* reason, const char* file, int line, int gsl_errno
) {
 throw formatted_exception("gsl: %2%:%3%: ERROR: %1%, errcode = %4%")
  << reason << file << line << gsl_errno;
};

int main(int argc, char** argv) {
 try {
  // Set output floating point format
  std::cout.setf(std::ios::scientific | std::ios::right);
  std::cout.precision(5);
  std::cerr.setf(std::ios::scientific | std::ios::right);
  std::cerr.precision(5);

  // Parse command-line parameters
  boost::program_options::variables_map::const_iterator option;
  boost::program_options::variables_map options = parse_arguments(argc, argv);

#ifdef DEBUG
  // Initialize debug output
  if (options.find("debug") != options.end())
   debug.out = &std::cerr;
#endif

  // Initialize input
  iosubgroups* io;
  if (options.find("data") != options.end())
   io = new iosubgroups_abbn(options);
  else
   io = new iosubgroups_text(options);

  double maxerror = options["error"].as<double>();

  gsl_set_error_handler(_gsl_error_handler);

  // Process groups
  while (!io->done()) {
   try {
    double err;
    std::vector<subgroup> subgroups;
    group_data gd = io->get();
    try {
     if (!gd.K)
      subgroups = compute_subgroups(
        gd.total, gd.partials, gd.knots, maxerror, &err, gd.ignore
      );
     else
      subgroups = compute_subgroups_K(
        gd.total, gd.partials, gd.knots, gd.K, maxerror, &err, gd.ignore
      );
    } catch (subgroups::bad_accuracy& e) {
     subgroups = e.result;
     err = e.error;
    };

    io->set(gd, subgroups);
    std::cerr 
     << std::setw(3) << gd.group
     << (gd.ignore >= 0 ? '*' : ' ') << ' '
     << std::setw(2) << subgroups.size() << ' '
     << std::setw(12) << err
     << (err < maxerror ? "\n" : " !\n");
   } catch (std::exception& e) {
    std::cerr << e.what() << '\n';
   };
  };

  delete io;
  return 0;
 } catch (std::exception& e) {
  std::cerr << e.what() << '\n';
  return 1;
 };
};
