#include "pade.hpp"
#include "formatted_exception.hpp"

static inline double sqr(double x) { 
 return x*x;
};

pade::pade(double z, double c_2, double c_1) {
 nominator.resize(1);
 nominator[0] = c_1;
 denominator.resize(2);
 denominator[0] = -z-c_2/c_1;
 denominator[1] = 1;
};

pade::pade(double z, double c_2, double c_1, double c0, double c1) {
 double d = c_1*c_1 + c0*c_2;
 if (d == 0)
  throw formatted_exception(
    "pade::pade(%1%, %2%, %3%, %4%, %5%): degenerate approximation"
  ) << z << c_2 << c_1 << c0 << c1;

 nominator.resize(2);
 denominator.resize(3);

 denominator[0] = 1;
 denominator[1] = (c0*c_1 - c1*c_2) / d;
 denominator[2] = (c0*c0 + c_1*c1)  / d;
 nominator[0]   = c0;
 nominator[1]   = c_1 * denominator[2];

 nominator.nshift(-z);
 denominator.nshift(-z);
};

// returns (z + alpha) * current + beta * (z - zcurrent)^2 * previous
polynome transform_polynome(
  const polynome& previous, const polynome& current, 
  double alpha, double beta, double zcurrent
) {
 double zsqr = zcurrent * zcurrent;
 size_t n = current.power()+1;
 polynome result(n+1);
 result[0] = alpha * current[0] + beta * zsqr * previous[0];
 result[1] = alpha * current[1] + current[0] 
               - 2 * beta * zcurrent * previous[0] 
               + (previous.size() > 1 ? beta * zsqr * previous[1] : 0);
 for (size_t k = 2; k < n-1; ++k)
  result[k] = alpha * current[k] + current[k-1]
              + beta * previous[k-2] - 2 * beta * zcurrent * previous[k-1]
              + beta * zsqr * previous[k];
 if (n > 2)
  result[n-1] = alpha * current[n-1] + current[n-2] 
                + beta * previous[n-3] - 2 * beta * zcurrent * previous[n-2];
 result[n] = current[n-1] + beta * previous[n-2];
 return result;
};

pade pade::add_point(
  double zcurrent, double znext, 
  double c0, double c1, 
  const pade& previous, const pade& current
) {
 polynome p0 = previous.nominator.shift(znext);
 p0.resize(2);
 polynome q0 = previous.denominator.shift(znext);
 q0.resize(2);
 polynome p1 = current.nominator.shift(znext);
 p1.resize(2);
 polynome q1 = current.denominator.shift(znext);
 q1.resize(2);

 double b = p1[0] - c0*q1[0];
 double a = sqr(znext-zcurrent) * (p0[0] - c0*q0[0]);
 double d 
  = b * (znext-zcurrent)
    * (-2*(c0*q0[0] - p0[0]) + (znext-zcurrent)*(p0[1] - c0*q0[1] - c1*q0[0]))
    - a * (p1[1] - c0*q1[1] - c1*q1[0]);
 a *=  b/d;
 b *= -b/d;

 pade result;
 result.nominator = transform_polynome(
   previous.nominator, current.nominator, a-znext, b, zcurrent
 );
 result.denominator = transform_polynome(
   previous.denominator, current.denominator, a-znext, b, zcurrent
 );

 return result;
};

double pade::operator()(double x) {
 return nominator(x) / denominator(x);
};

#ifdef DEBUG
std::ostream& operator<<(std::ostream& s, const pade& p) {
 return s << '(' << p.nominator << ") / (" << p.denominator << ')';
};
#endif
