#include "polynome.hpp"
#include <stdexcept>

#include <math.h>

#include <gsl/gsl_poly.h>
#include <gsl/gsl_errno.h>

void divide(
  const polynome& divident, const polynome& divisor,
  polynome& ratio, polynome& remainder
) {
 remainder = divident;

 if (divident.size() < divisor.size()) {
  ratio.resize(1, 0);
  ratio.front() = 0;
  return;
 };

 ratio.resize(divident.size()-divisor.size()+1);
 for (size_t i = divident.size(); i-- >= divisor.size();) {
  ratio[i+1-divisor.size()] = remainder[i]/divisor.back();
  double r = remainder[i] / divisor.back();
  for (size_t j = i-divisor.size()+1; j < i; ++j)
   remainder[j] -= r * divisor[divisor.size()-1 + j-i];
 };
 remainder.resize(divisor.size()-1);
 remainder.normalize();
};

polynome::polynome() {};

polynome::polynome(double a): std::vector<double>(1, a) {};

polynome::polynome(size_t size, double value)
 : std::vector<double>(size, value) {};

polynome::polynome(const std::vector<double>& v): std::vector<double>(v) {
 if (size() == 0) resize(1, 0);
};

polynome::polynome(const polynome& p): std::vector<double>(p) {};

polynome polynome::operator-() const {
 polynome result(*this);
 for (iterator i = result.begin(); i != result.end(); ++i)
  *i = -*i;
 return result;
};

polynome polynome::operator+(double a) const {
 return polynome(*this) += a;
};

polynome polynome::operator-(double a) const {
 return polynome(*this) -= a;
};

polynome polynome::operator*(double a) const {
 return polynome(*this) *= a;
};

polynome polynome::operator/(double a) const {
 return polynome(*this) /= a;
};

polynome polynome::operator+(const polynome& p) const {
 return polynome(*this) += p;
};

polynome polynome::operator-(const polynome& p) const {
 return polynome(*this) -= p;
};

polynome polynome::operator*(const polynome& p) const {
 if (p == 0) return p;
 polynome result(size()+p.size()-1, 0);
 for (size_t k = 0; k < result.size(); ++k)
  for (size_t n = k < p.size() ? 0 : k-p.size()+1; 
    n <= (k < size() ? k : size()-1); ++n)
   result[k] += (*this)[n]*p[k-n];
 return result;
};

polynome polynome::operator/(const polynome& p) const {
 polynome result;
 polynome tmp;
 divide(*this, p, result, tmp);
 return result;
};

polynome polynome::operator%(const polynome& p) const {
 polynome result;
 polynome tmp;
 divide(*this, p, tmp, result);
 return result;
};

polynome& polynome::operator+=(double a) {
 front() += a;
 return *this;
};

polynome& polynome::operator-=(double a) {
 front() -= a;
 return *this;
};

polynome& polynome::operator*=(double a) {
 if (a == 0) {
  resize(1);
  front() = 0;
  return *this;
 };
 for (iterator i = begin(); i != end(); ++i)
  *i *= a;
 return *this;
};

polynome& polynome::operator/=(double a) {
 for (iterator i = begin(); i != end(); ++i)
  *i /= a;
 return *this;
};

polynome& polynome::operator+=(const polynome& p) {
 iterator i = begin();
 const_iterator j = p.begin();
 while (i != end() && j != p.end())
  *i++ += *j++;
 while (j != p.end())
  push_back(*j++);
 normalize();
 return *this;
};

polynome& polynome::operator-=(const polynome& p) {
 iterator i = begin();
 const_iterator j = p.begin();
 while (i != end() && j != p.end())
  *i++ -= *j++;
 while (j != p.end())
  push_back(-*j++);
 normalize();
 return *this;
};

polynome& polynome::operator*=(const polynome& p) {
 return *this = *this * p;
};

polynome& polynome::operator/=(const polynome& p) {
 return *this = *this / p;
};

polynome& polynome::operator%=(const polynome& p) {
 return *this = *this % p;
};

bool polynome::operator==(double a) const {
 return size() == 1 && front() == a;
};

bool polynome::operator!=(double a) const {
 return size() != 1 || front() != a;
};

bool polynome::operator==(const polynome& p) const {
 return !((*this) != p);
};

bool polynome::operator!=(const polynome& p) const {
 if (size() != p.size())
  return true;
 const_iterator i = begin();
 const_iterator j = p.begin();
 while (i != end())
  if (*i++ != *j++)
   return true;
 return false;
};

double polynome::operator()(double x) const {
 double z = 1;
 double result = 0;
 for (const_iterator i = begin(); i != end(); ++i) {
  result += *i * z;
  z *= x;
 };
 return result;
};

std::vector<double> polynome::roots() const {
 int i = size()-1;
 while ((*this)[i] == 0) --i;
 switch (i) {
  case 0: return std::vector<double>();
  case 1: return std::vector<double>(1, -(*this)[0]/(*this)[1]);
  case 2: {
    std::vector<double> result(2);
    switch (gsl_poly_solve_quadratic(
       (*this)[2], (*this)[1], (*this)[0],
       &result[0], &result[1]
      )) {
     case 0: return std::vector<double>();
     case 1: result.resize(1);
     default: return result;
    };
   };
  case 3: {
   std::vector<double> result(3);
   if (gsl_poly_solve_cubic(
      (*this)[2]/(*this)[3], 
      (*this)[1]/(*this)[3], 
      (*this)[0]/(*this)[3],
      &result[0], &result[1], &result[2]) == 1)
    result.resize(1);
   // For the following polynome GSL returns nan values 
   // (untested in the precision shown):
   // 1.96581e+04+9.30186e+04*x+7.36349e+04*x**2-7.00116e-04*x**3
   if (isnan(result[0])) return std::vector<double>();
   return result;
  };
  default:
   gsl_poly_complex_workspace* w = gsl_poly_complex_workspace_alloc(size());
   if (!w) 
    throw std::runtime_error("polynome::roots: error during memory allocation");
   int flag;
   double* z = new double[2*(size()-1)];
   try {
    flag = gsl_poly_complex_solve(&(*this)[0], size(), w, z);

    gsl_poly_complex_workspace_free(w);

   } catch (...) {
    gsl_poly_complex_workspace_free(w);
    delete[] z;
    throw;
   };

   try {

    if (flag != GSL_SUCCESS)
     // I've just read this stuff in GSL documentation :)
     // (though there is some explanation of the algorithm on wikipedia)
     throw 
      std::runtime_error(
	"Unable to solve equation: balanced QR-reduction of companion "
        "matrix does not converge"
      );
    std::vector<double> result(size()-1);
    for (size_t i = 0; i < size()-1; ++i) {
     if (z[2*i+1] != 0)  
      // Complex roots are correct solution, but we can't use them
      return std::vector<double>();
     result[i] = z[2*i];
    };

    delete[] z;
    return result;

   } catch (...) {
    delete[] z;
    throw;
   };
 };
};

void polynome::normalize() {
 if (empty()) resize(1, 0);
 else while (size() > 1 && back() == 0) pop_back();
};

polynome operator+(double x, const polynome& p) {
 return p+x;
};

polynome operator-(double x, const polynome& p) {
 polynome result(p);
 for (polynome::iterator i = result.begin(); i != result.end(); ++i)
  *i = -*i;
 result.front() += x;
 return result;
};

polynome operator*(double x, const polynome& p) {
 return p*x;
};

bool operator==(double x, const polynome& p) {
 return p == x;
};

bool operator!=(double x, const polynome& p) {
 return p != x;
};

polynome& polynome::nderivate() {
 for (size_t i = 1; i < size(); ++i)
  (*this)[i-1] = (*this)[i]*i;
 if (size() > 1) pop_back();
 else front() = 0;
 return *this;
};

polynome polynome::derivate() const {
 return polynome(*this).nderivate();
};

polynome& polynome::nshift(double dx) {
 std::vector<double> c(size(), 1);
 for (size_t m = 0; m < size(); ++m) {
  double dxm = dx;
  for (size_t k = m+1; k < size(); ++k) {
   (*this)[m] += (*this)[k] * c[k] * dxm;
   dxm *= dx;
   c[k] *= (double)(k-m)/(m+1);
  };
 };
 return *this;
};

polynome polynome::shift(double dx) const {
 return polynome(*this).nshift(dx);
};

#ifdef DEBUG
std::ostream& operator<<(std::ostream& s, const polynome& v) {
 if (v.empty()) return s;
 s << v[0];
 if (v.size() >= 2) {
  std::ios::fmtflags flags = s.setf(std::ios::showpos);
  s << v[1] << "*x";
  for (size_t i = 2; i < v.size(); ++i)
   s << v[i] << "*x**" << i;
  s.flags(flags);
 };
 return s;
};
#endif
