#include <algorithm>

#include <boost/format.hpp>

#include <string.h>
#include <math.h>

#include <gsl/gsl_vector.h>
#include <gsl/gsl_permute.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_multimin.h>

#include "subgroups.hpp"
#include "pade.hpp"
#include "gsl.hpp"

#ifdef DEBUG
#include "debug.hpp"
#endif

namespace subgroups {

static std::ostream& operator<<(std::ostream& s, const std::vector<double>& v) {
 s << '[';
 for (std::vector<double>::const_iterator i = v.begin(); i != v.end(); ++i)
  s << ' ' << *i;
 s << " ]";
 return s;
};

bad_accuracy::bad_accuracy(double error, const std::vector<subgroup>& result)
  : error(error), result(result) {
 boost::format f(
    "Could not achieve the required accuracy. %1% is the best result"
 );
 f % error;
 message = new char[f.size()+1];
 strcpy(message, f.str().c_str());
};

pade compute_pade(
  double total0, 
  const std::vector<knot>& knots, 
  const std::vector<size_t>& points
) {
 pade p0(0, -total0, 1);
 pade p1(
   knots[points[0]].dilution, 
   -total0 - knots[points[0]].dilution, 
   1, 
   knots[points[0]].total1, 
   -knots[points[0]].total2
 );
 pade* P0 = &p0;
 pade* P1 = &p1;
 for (size_t i = 1; i < points.size(); ++i) {
  *P0 = pade::add_point(
          knots[points[i-1]].dilution, knots[points[i]].dilution,
          knots[points[i]].total1, -knots[points[i]].total2,
          *P0, *P1
        );
  std::swap(P0, P1);
 };
 return *P1;
};

class padeCache {
 public:
  pade cache[2];
  double total0;
  double last_dilution;

  padeCache(
    double total0,
    const std::vector<std::vector<knot>::const_iterator>& points
  ): total0(total0) {
   fill(points); 
  };
  pade operator()(const knot& k);
  void fill(const std::vector<std::vector<knot>::const_iterator>& points);
};

void padeCache::fill(
  const std::vector<std::vector<knot>::const_iterator>& points
) {
 cache[0] = pade(0, -total0, 1);
 cache[1] = pade(
   points.back()->dilution,
   -total0 - points.back()->dilution,
   1,
   points.back()->total1,
   -points.back()->total2
 );

 pade* p0 = &cache[0];
 pade* p1 = &cache[1];
 for (std::vector<std::vector<knot>::const_iterator>::const_iterator p
       = points.end()-2; p != points.begin(); --p) {
  *p0 = pade::add_point(
    (*(p+1))->dilution, (*p)->dilution, (*p)->total1, -(*p)->total2, *p0, *p1
  );
  std::swap(p0, p1);
 };
 if (p0 != &cache[0])
  std::swap(cache[0], cache[1]);

 last_dilution = points[1]->dilution;
};

pade padeCache::operator()(const knot& k) {
 return pade::add_point(
   last_dilution, k.dilution, k.total1, -k.total2, cache[0], cache[1]
 );
};

std::vector<subgroup> compute_total(const pade& p) {
 std::vector<double> s = p.denominator.roots();
 std::vector<double> a;

 size_t K = s.size();

 if (K+1 < p.denominator.size()) // complex roots
  return std::vector<subgroup>();

 std::sort(s.begin(), s.end());
 {
  std::vector<double>::const_iterator r1 = s.begin();
  std::vector<double>::const_iterator r0 = r1++;
  for (; r1 != s.end(); ++r1, ++r0)
   if (*r1 == *r0) return std::vector<subgroup>(); // equal roots 
                                                   // (as useless as complex)
 };

 // TODO: simplify
 {
  std::vector<double> r = p.nominator.roots();
  for (std::vector<double>::const_iterator x = r.begin(); x != r.end(); ++x)
   for (std::vector<double>::const_iterator y = s.begin(); y != s.end(); ++y)
    if (fabs(*x / *y) < 1E-6) return std::vector<subgroup>();
 };

 double* matrix = new double[K*K];

 try {
  for (size_t i = 0; i < K; ++i) {
   matrix[(K-1)*K+i] = p.denominator.back();
   for (size_t j = K-1; j > 0; --j)
    matrix[(j-1)*K+i] = p.denominator[j] + matrix[j*K+i]*s[i];
  };

//  print_system(debug, matrix, &p.nominator[0], K);

  a = system_solve(matrix, &p.nominator[0], K);

  delete[] matrix;
 } catch (system_degenerated_error&) {
  delete[] matrix;
  return std::vector<subgroup>();
 } catch (...) {
  delete[] matrix;
  throw;
 };

 std::vector<subgroup> result(K);
 for (size_t i = 0; i < K; ++i) {
  result[i].total = -s[i];
  result[i].probability = a[i];
 };

 return result;
};

std::vector<subgroup>& compute_partials(
  std::vector<subgroup>& subgroups,
  const std::vector<double>& partials,
  const std::vector<knot>& knots
) {
 size_t K = subgroups.size();
 size_t P = partials.size();

 size_t u = 0;
 for (size_t x = 1; x < P; ++x)
  if (partials[x] > partials[u])
   u = x;

 double total = 0;
 for (std::vector<double>::const_iterator x = partials.begin(); 
   x != partials.end(); ++x)
  total += *x;

 std::vector<double> s;

 size_t N = (K-1)*(P-1);
 double* matrix = new double[N*N];
 double* rhs = new double[N];
 double* m = matrix;
 double* r = rhs;

 try {

  for (size_t v = 0; v < P; ++v) {
   if (v == u) continue;
   double d = total - partials[u] + partials[v];
   for (size_t n = 0; n < K-1; ++n) {
    for (size_t x = 0; x < P; ++x) {
     if (x == u) continue;
     for (size_t k = 0; k < K-1; ++k) {
      *m = 0;
      for (std::vector<knot>::const_iterator i = knots.begin(); 
        i != knots.end(); ++i) {
       double sKi = 1./(subgroups.back().total + i->dilution);
       *m += (1./(subgroups[k].total + i->dilution) - sKi)
           * (1./(subgroups[n].total + i->dilution) - sKi);
      };
      if (x == v) *m *= 2;
      *m++ *= subgroups[k].probability;
     };
    };
    *r = 0;
    for (std::vector<knot>::const_iterator i = knots.begin();
      i != knots.end(); ++i) {
     double t = 0;
     for (size_t x = 0; x < P; ++x) if (x != u) t += i->partial[x];
     double sKi = 1./(subgroups.back().total + i->dilution);
     *r += (t + i->partial[v] - d*sKi)
         * (1./(subgroups[n].total + i->dilution) - sKi);
    };
    ++r;
   };
  };

  s = system_solve(matrix, rhs, N);

  delete[] matrix;
  delete[] rhs;
 } catch (...) {
  delete[] matrix;
  delete[] rhs;
  throw;
 };

 for (std::vector<subgroup>::iterator k = subgroups.begin();
   k != subgroups.end(); ++k)
  k->partial.resize(P);

 std::vector<double>::const_iterator i = s.begin();
 for (size_t v = 0; v < P; ++v) {
  if (v == u) continue;
  double& z = subgroups.back().partial[v];
  z = partials[v];
  for (size_t n = 0; n < K-1; ++n) {
   subgroups[n].partial[v] = *i;
   z -= subgroups[n].probability * *i++;
  };
  z /= subgroups.back().probability;
 };

 for (std::vector<subgroup>::iterator k = subgroups.begin();
   k != subgroups.end(); ++k) {
  k->partial[u] = k->total;
  for (size_t x = 0; x < P; ++x) {
   if (x == u) continue;
   k->partial[u] -= k->partial[x];
  };
 };

 return subgroups;
};

std::vector<subgroup> compute_subgroups(
  const pade& p,
  const std::vector<double>& partials,
  const std::vector<knot>& knots
) {
 std::vector<subgroup> result = compute_total(p);
 if (result.empty()) return result;
 compute_partials(result, partials, knots);
 return result;
};

std::vector<knot>::const_iterator knots_begin;

static std::ostream& operator<<(
  std::ostream& s, const std::vector<std::vector<knot>::const_iterator>& points
) {
 for (std::vector<std::vector<knot>::const_iterator>::const_iterator p
   = points.begin(); p != points.end(); ++p)
  s << (*p - knots_begin) << ' ';
 return s;
};

std::vector<knot>::const_iterator increment(
  std::vector<std::vector<knot>::const_iterator>& points
) {
 do ++points.front();
 while (find(points.begin()+1, points.end(), points.front()) != points.end());
 return points.front();
};

bool positive(const std::vector<subgroup>& subgroups) {
 for (std::vector<subgroup>::const_iterator s = subgroups.begin(); 
   s != subgroups.end(); ++s) {
  if (s->probability < 0 || s->total < 0) return false;
  for (std::vector<double>::const_iterator x = s->partial.begin();
    x != s->partial.end(); ++x)
   if (*x < 0) return false;
 };
 return true;
};

double minimize_total_error(
  std::vector<subgroup>& subgroups,
  const std::vector<knot>& knots,
  double maxerror,
  int ignore = -1
);

double total_error(const std::vector<subgroup>&, const std::vector<knot>&);
double partial_error(
  const std::vector<subgroup>&, const std::vector<knot>&, int ignore = -1
);

struct subgroups_error {
 std::vector<subgroup> subgroups;
 const std::vector<knot>& knots;
 double total_error;
 double partial_error;
 bool positive;
 double maxerror;
 int ignore;

 subgroups_error(
   const std::vector<knot>& knots,
   double maxerror,
   int ignore = -1
 ): knots(knots), maxerror(maxerror), ignore(ignore) { test(); };

 subgroups_error(
   std::vector<subgroup> subgroups, 
   const std::vector<knot>& knots, 
   double maxerror,
   int ignore = -1
 ): subgroups(subgroups), knots(knots), maxerror(maxerror), ignore(ignore) {
  test();
 };

 subgroups_error& operator=(const subgroups_error& e) {
  subgroups = e.subgroups;
  total_error = e.total_error;
  partial_error = e.partial_error;
  positive = e.positive;
  maxerror = e.maxerror;
  ignore = e.ignore;
  return *this;
 };

 subgroups_error& operator=(const std::vector<subgroup>& sgs) {
  subgroups = sgs;
  test();
  return *this;
 };

 inline double error() const {
  return fmax(total_error, partial_error);
 };

 inline bool enough() const {
  return error() < maxerror;
 };

 bool valid() const {
  for (std::vector<subgroup>::const_iterator s = subgroups.begin(); 
    s != subgroups.end(); ++s) {
   if (!isfinite(s->probability) || !isfinite(s->total))
    return false;
   for (std::vector<double>::const_iterator x = s->partial.begin();
     x != s->partial.end(); ++x)
    if (!isfinite(*x)) return false;
  };
  return true;
 };

 bool degenerated() const {
  for (std::vector<subgroup>::const_iterator s = subgroups.begin();
    s != subgroups.end(); ++s)
   // TODO: let user change this value
   if (s->probability < 1E-3)
    return true;
  return false;
 };

 double minimize(const std::vector<knot>& knots) {
#ifdef DEBUG
  debug << subgroups << '\n';
#endif
  minimize_total_error(subgroups, knots, maxerror, ignore);
#ifdef DEBUG
  debug << subgroups << '\n';
#endif
  test();
  return error();
 };

 void test() {
  if (subgroups.empty() || !valid()) {
   total_error = partial_error = 1;
   positive = false;
  } else {
   total_error = subgroups::total_error(subgroups, knots);
   partial_error = subgroups::partial_error(subgroups, knots, ignore);
   positive = subgroups::positive(subgroups);
  };
 };

 void clear() {
  subgroups.clear();
  test();
 };
};

std::ostream& operator<<(std::ostream& s, const subgroups_error& e) {
 return s << "total_error = " << e.total_error
  << ", partial_error = " << e.partial_error
  << ", positive = " << e.positive
  << ", enough = " << e.enough();
};

void test_line(
  double total0,
  const std::vector<double>& partials,
  std::vector<std::vector<knot>::const_iterator>& points,
  const std::vector<knot>& knots,
  subgroups_error& result
) {
 padeCache pade_cache(total0, points);
 
 std::vector<knot>::const_iterator result_point = points.front();
 points.front() = knots.begin();
 --points.front();
 increment(points); // FIXME
 subgroups_error current(result);

 if (result.error() < 0) {
  do {
   try {
    result = compute_subgroups(pade_cache(*points.front()), partials, knots);
   } catch (std::exception& e) {
#ifdef DEBUG
    debug << e.what() << '\n';
#endif
   };
#ifdef DEBUG
   debug << "test_line: points = " << points
    << ", result: " << result << '\n';
#endif
   increment(points);
  } while (result.subgroups.empty() && points.front() != knots.end());
 };

 for (; !(result.enough() && result.positive)
        && points.front() != knots.end(); 
   increment(points)) {
  try {
   current = compute_subgroups(pade_cache(*points.front()), partials, knots);
   if (current.subgroups.empty()) continue;
   // TODO: simplify
   if (current.total_error < result.total_error 
       && (result.enough() ? current.enough() : true)
       || current.enough() && current.positive) {
    result = current;
    result_point = points.front();
#ifdef DEBUG
    debug << "test_line: points = " << points 
     << ", result: " << result << '\n' << result.subgroups;
#endif
   } 
#ifdef DEBUG
   else
    debug << "test_line: points = " << points
     << ", current: " << current << '\n';
#endif
  } catch (std::exception& e) {
#ifdef DEBUG
   debug << e.what() << '\n';
#endif
  };
 };

 points.front() = result_point;
};

void test_volume(
  double total0, 
  const std::vector<double>& partials,
  const std::vector<knot>& knots,
  size_t K,
  subgroups_error& result
) {
 result.clear();
 std::vector<std::vector<knot>::const_iterator> points(K-1);
 {
  knots_begin = knots.begin();
  std::vector<knot>::const_iterator k = knots.begin();
  std::vector<std::vector<knot>::const_iterator>::reverse_iterator p 
   = points.rbegin();
  while (p != points.rend())
   *p++ = k++;
 };

 size_t trying = 0;
 do {
  std::vector<knot>::const_iterator current = points.front();
  try {
   test_line(total0, partials, points, knots, result);
  } catch (std::exception& e) {
#ifdef DEBUG
   debug << e.what() << '\n';
#endif
  }
  if (current == points.front()) ++trying;
  else trying = 1;
#ifdef DEBUG
  debug << "test_volume: points = " << points 
   << ", result: " << result << '\n';
#endif
  std::rotate(points.begin(), points.begin()+1, points.end());
 } while (!(result.enough() && result.positive) && trying < points.size());
};

static inline double sqr(double x) {
 return x*x;
};

double mean(const std::vector<subgroup>& subgroups) {
 double result = 0;
 for (std::vector<subgroup>::const_iterator s = subgroups.begin();
   s != subgroups.end(); ++s)
  result += s->probability * s->total;
 return result;
};

double mean(const std::vector<subgroup>& subgroups, size_t x) {
 double result = 0;
 for (std::vector<subgroup>::const_iterator s = subgroups.begin();
   s != subgroups.end(); ++s)
  result += s->probability * s->partial[x];
 return result;
};

double momenta1(const std::vector<subgroup>& subgroups, double dilution) {
 double result = 0;
 for (std::vector<subgroup>::const_iterator s = subgroups.begin();
   s != subgroups.end(); ++s)
  result += s->probability / (s->total + dilution);
 return result;
};

double momenta2(const std::vector<subgroup>& subgroups, double dilution) {
 double result = 0;
 for (std::vector<subgroup>::const_iterator s = subgroups.begin();
   s != subgroups.end(); ++s)
  result += s->probability / sqr(s->total + dilution);
 return result;
};

double momenta1x(
  const std::vector<subgroup>& subgroups, double dilution, size_t x
) {
 double result = 0;
 for (std::vector<subgroup>::const_iterator s = subgroups.begin();
   s != subgroups.end(); ++s)
  result += s->probability * s->partial[x] / (s->total + dilution);
 return result;
};

double relative_error(double x, double y) {
 return fabs(y/x-1);
};

double blocked_xsection(const std::vector<subgroup>& subgroups, double dilution) {
 double m1 = 0;
 double m2 = 0;
 for (std::vector<subgroup>::const_iterator s = subgroups.begin(); 
   s != subgroups.end(); ++s) {
  m1 += s->probability / (s->total + dilution);
  m2 += s->probability / sqr(s->total + dilution);
 };
 return m1/m2 - dilution;
};

double blocked_xsection(
  const std::vector<subgroup>& subgroups, double dilution, size_t x
) {
 double m1 = 0;
 double mx = 0;
 for (std::vector<subgroup>::const_iterator s = subgroups.begin();
   s != subgroups.end(); ++s) {
  m1 += s->probability / (s->total + dilution);
  mx += s->probability * s->partial[x] / (s->total + dilution);
 };
 return mx/m1;
};

double total_error(
  const std::vector<subgroup>& subgroups,
  const std::vector<knot>& knots
) {
 double result = 0;
 for (std::vector<knot>::const_iterator k = knots.begin(); k != knots.end(); 
   ++k) {
  result = fmax(
    result,
    relative_error(
     k->total1 / k->total2 - k->dilution,
     blocked_xsection(subgroups, k->dilution)
    )
  );
 };
 return result;
};

double partial_error(
  const std::vector<subgroup>& subgroups,
  const std::vector<knot>& knots,
  int ignore
) {
 double result = 0;
 for (std::vector<knot>::const_iterator k = knots.begin(); k != knots.end();
   ++k) {
  for (int x = 0; x < subgroups.front().partial.size(); ++x) {
   if (x == ignore) continue;
   result = fmax(
     result,
     relative_error(
      k->partial[x] / k->total1,
      blocked_xsection(subgroups, k->dilution, x)
     )
   );
  };
 };
 return result;
};

inline double error(
  const std::vector<subgroup>& subgroups,
  const std::vector<knot>& knots,
  int ignore = -1
) {
 return fmax(
   total_error(subgroups, knots), partial_error(subgroups, knots, ignore)
 );
};

double total_sqr_error(
  const std::vector<subgroup>& subgroups,
  const std::vector<knot>& knots
) {
 double result = 0;
 for (std::vector<knot>::const_iterator k = knots.begin(); 
   k != knots.end(); ++k)
  result += sqr(
   relative_error(
    k->total1 / k->total2 - k->dilution,
    blocked_xsection(subgroups, k->dilution)
   )
  );
 return result;
};

double partial_sqr_error(
  const std::vector<subgroup>& subgroups,
  const std::vector<knot>& knots
) {
 double result = 0;
 for (size_t x = 0; x < subgroups.front().partial.size(); ++x)
  for (std::vector<knot>::const_iterator k = knots.begin();
    k != knots.end(); ++k)
   result += sqr(
    relative_error(
     k->partial[x] / k->total1,
     blocked_xsection(subgroups, k->dilution, x)
    )
   );
 return result;
};

inline double sqr_error(
  const std::vector<subgroup>& subgroups,
  const std::vector<knot>& knots
) {
 return total_sqr_error(subgroups, knots) + partial_sqr_error(subgroups, knots);
};

inline double cot(double x) {
 return tan(M_PI/2 - x);
};

inline double acot(double x) {
 return M_PI/2 - atan(x);
};

// Ensures that x lies in (0; b)
inline double bound(double x, double b) {
 const double eps = 1E-2;
#ifdef DEBUG
 debug << "bound(" << x << ", " << b << ")\n";
#endif
 if (x <= 0)
  return eps*b;
 if (x >= b)
  return (1-eps)*b;
 return x;
};
// TODO; research alternative transformations
// Maps x from [-inf; inf] to [0; bound]
inline double transform(double x, double bound) {
 return cot(x*M_PI/bound);
};
// Maps x from [0; bound] to [-inf; inf]
inline double untransform(double x, double bound) {
 return bound/M_PI * acot(x);
};

void simplex_transform(gsl_vector* x, double bound) {
 double sum = 0;
 for (size_t i = 0; i < x->size; ++i)
  sum += gsl_vector_get(x, i);
 for (size_t i = 0; i < x->size; ++i) {
  double xi = gsl_vector_get(x, i);
  gsl_vector_set(x, i, transform(xi, bound - (sum - xi)));
 };
};

void simplex_untransform(gsl_vector* x, double bound) {
 // TODO: investigate matrix properties
 gsl_matrix* m = gsl_matrix_alloc(x->size, x->size);
 gsl_matrix_set_all(m, 1);
 for (size_t i = 0; i < x->size; ++i)
  gsl_matrix_set(m, i, i, M_PI/acot(gsl_vector_get(x, i)));

 gsl_vector* rhs = gsl_vector_alloc(x->size);
 gsl_vector_set_all(rhs, bound);

 gsl_permutation* p = gsl_permutation_alloc(x->size);
 int signum;

 gsl_linalg_LU_decomp(m, p, &signum);
 gsl_linalg_LU_solve(m, p, rhs, x);

 gsl_permutation_free(p);
 gsl_vector_free(rhs);
 gsl_matrix_free(m);
};

gsl_vector* apply_transform(
  gsl_vector* x, 
  double total, 
  const std::vector<double>& partials,
  void (*transform)(gsl_vector*, double)
) {
 size_t n = x->size/(partials.size()+1);
 transform(&gsl_vector_subvector(x, 0, n).vector, 1);
 transform(&gsl_vector_subvector(x, n, n).vector, total);
 for (size_t m = 0; m < partials.size()-1; ++m)
  transform(&gsl_vector_subvector(x, n*(m+2), n).vector, partials[m]);
 return x;
};

inline void escape_boundary(gsl_vector* x, double bound, double proximity) {
 const double eps = 1E-5;
 double margin = bound*eps;
 if (proximity >= margin) return;
#ifdef DEBUG
 debug << "escape_boundary(x, " << bound << ", " << proximity << ")\n";
#endif
 margin = (margin - proximity) / x->size;
 for (size_t i = 0; i < x->size; ++i)
  gsl_vector_set(x, i, gsl_vector_get(x, i) - margin);
};

gsl_vector* subgroups_to_gsl_vector(const std::vector<subgroup>& subgroups) {
 size_t n = subgroups.size()-1;
 size_t m = subgroups.front().partial.size();
 gsl_vector* result = gsl_vector_alloc(n * (m+1));
 double total = mean(subgroups);
 std::vector<double> partials(m);

 size_t u = 0;
 for (size_t x = 0; x < m; ++x) {
  partials[x] = mean(subgroups, x);
  if (partials[x] > partials[u]) u = x;
 };

 double dp = 1;
 for (size_t k = 0; k < n; ++k) {
  double p = bound(subgroups[k].probability, 1);
  dp -= p;
  gsl_vector_set(result, k, p);
 };
 escape_boundary(&gsl_vector_subvector(result, 0, n).vector, 1, dp);
 double dt = total;
 for (size_t k = 0; k < n; ++k) {
  double p = bound(gsl_vector_get(result, k) * subgroups[k].total, total);
  dt -= p;
  gsl_vector_set(result, k+n, p);
 };
 escape_boundary(&gsl_vector_subvector(result, n, n).vector, total, dt);
 for (size_t x = 0; x < m; ++x) {
  if (x == u) continue;
  size_t index = (x + (x > u ? 1 : 2)) * n;
  double dpx = partials[x];
  for (size_t k = 0; k < n; ++k) {
   double p 
    = bound(gsl_vector_get(result, k) * subgroups[k].partial[x], partials[x]);
   dpx -= p;
   gsl_vector_set(result, k + index, p);
  };
  escape_boundary(
    &gsl_vector_subvector(result, index, n).vector, partials[x], dpx
  );
 };

// partials(m-1) = mean(subgroup, m-1); // isn't used in apply_transform anyway
 apply_transform(result, total, partials, simplex_transform);
 return result;
};

std::vector<subgroup> gsl_vector_to_subgroups(
  const gsl_vector* x, double total, const std::vector<double>& partials
) {
 gsl_vector* v = gsl_vector_alloc(x->size);
 gsl_vector_memcpy(v, x);
 apply_transform(v, total, partials, simplex_untransform);

 size_t u = 0;
 for (size_t x = 1; x < partials.size(); ++x)
  if (partials[x] > partials[u]) u = x;

 size_t n = v->size/(partials.size()+1);
 std::vector<subgroup> result(n + 1);
 double t = total;
 double p = 1;
 for (size_t k = 0; k < n; ++k) {
  result[k].probability = gsl_vector_get(v, k);
  p -= result[k].probability;
  t -= result[k].total = gsl_vector_get(v, k+n);
  result[k].total /= result[k].probability;
  result[k].partial.resize(partials.size());
 };
 result.back().probability = p;
 result.back().total = t/p;
 result.back().partial.resize(partials.size());
 for (size_t x = 0; x < partials.size(); ++x) {
  if (x == u) continue;
  t = partials[x];
  for (size_t k = 0; k < n; ++k) {
   t -= result[k].partial[x] = gsl_vector_get(v, k + (x + (x > u ? 1 : 2))*n);
   result[k].partial[x] /= result[k].probability;
  };
  result.back().partial[x] = t/p;
 };
 for (std::vector<subgroup>::iterator s = result.begin(); s != result.end(); 
   ++s) {
  s->partial[u] = s->total;
  for (size_t x = 0; x < partials.size(); ++x)
   if (x != u) s->partial[u] -= s->partial[x];
 };

 gsl_vector_free(v);

 return result;
};

struct minimizer_parameters {
 double total;
 std::vector<double> partials;
 const std::vector<knot>& knots;
 double real_error;
 int ignore;

 minimizer_parameters(
   const std::vector<subgroup>& subgroups, 
   const std::vector<knot>& knots,
   int ignore = -1
 ): knots(knots), real_error(0), ignore(ignore) {
  total = mean(subgroups);
  partials.resize(subgroups.front().partial.size());
  for (size_t x = 0; x != partials.size(); ++x)
   partials[x] = mean(subgroups, x);
 };
};

double minimizer_function(const gsl_vector* x, minimizer_parameters* params) {
 std::vector<subgroup> subgroups = gsl_vector_to_subgroups(
   x, params->total, params->partials
 );
 params->real_error = error(subgroups, params->knots, params->ignore);
 return params->real_error;
// return sqr_error(subgroups, params->knots);
};

double minimize_total_error(
  std::vector<subgroup>& subgroups,
  const std::vector<knot>& knots,
  double maxerror,
  int ignore 
) {
#ifdef DEBUG
 debug << "minimize_total_error(" << subgroups.size() << ")\n";
#endif
 gsl_vector* x = subgroups_to_gsl_vector(subgroups);
 gsl_vector* step_size = gsl_vector_alloc(x->size);
 gsl_vector_set_all(step_size, 1);

 gsl_multimin_function F;
 F.n = x->size;
 F.f = (double (*)(const gsl_vector*, void*))&minimizer_function;
 minimizer_parameters params(subgroups, knots, ignore);
 F.params = &params;

#ifdef DEBUG
 debug << gsl_vector_to_subgroups(x, params.total, params.partials);
#endif
 
 gsl_multimin_fminimizer* minimizer = gsl_multimin_fminimizer_alloc(
   gsl_multimin_fminimizer_nmsimplex, x->size
 );
 gsl_multimin_fminimizer_set(minimizer, &F, x, step_size);

 double err = 0;

 try {
  static const size_t max_iteration = 100;
  size_t iteration = 0;
  double olderr = -1;

  do {
   if (gsl_multimin_fminimizer_iterate(minimizer)) break;
   err = gsl_multimin_fminimizer_minimum(minimizer);
   if (err != olderr) {
    olderr = err;
    iteration = 0;
   } else ++iteration;
#ifdef DEBUG
   debug << "best = " << err << ", " << params.real_error << '\n';
#endif
  } while (params.real_error >= maxerror && iteration < max_iteration);

  subgroups = gsl_vector_to_subgroups(
    gsl_multimin_fminimizer_x(minimizer), params.total, params.partials
  );

  gsl_multimin_fminimizer_free(minimizer);
  gsl_vector_free(step_size);
  gsl_vector_free(x);
 } catch (...) {
  gsl_multimin_fminimizer_free(minimizer);
  gsl_vector_free(step_size);
  gsl_vector_free(x);
  throw;
 };

 return err;
};

};

std::vector<subgroup> compute_subgroups0(
  double total, const std::vector<double>& partials
) {
 std::vector<subgroup> result(1);
 result[0].probability = 1;
 result[0].total = total;
 result[0].partial = partials;
 return result;
};

std::vector<subgroup> compute_subgroups1(
  double total, 
  const std::vector<double>& partials,
  const std::vector<subgroups::knot>& knots
) {
 subgroups::knot k = knots.front();
 return compute_subgroups(
   pade(k.dilution, -total - k.dilution, 1, k.total1, -k.total2), 
   partials, 
   knots
 );
};

std::vector<subgroup> compute_subgroups_K(
  double total,
  const std::vector<double>& partials,
  const std::vector<subgroups::knot>& knots,
  size_t K,
  double maxerror,
  double* error,
  int ignore
) {
 double total0 = 0;
 for (std::vector<double>::const_iterator x = partials.begin();
   x != partials.end(); ++x)
  total0 += *x;

 subgroups::subgroups_error result(knots, maxerror, ignore);
 switch (K) {
  case 1: 
   result.subgroups = compute_subgroups0(total0, partials);
   if (error) result.test();
   break;
  case 2: 
   result.subgroups = compute_subgroups1(total0, partials, knots);
   result.test();
   if (result.total_error < maxerror && !result.positive)
    result.minimize(knots);
   break;
  default:
   test_volume(total0, partials, knots, K, result);
   if (!(result.enough() && result.positive))
    result.minimize(knots);
 };

 if (error) *error = result.error();
 double nonresonant = total - total0;
 if (nonresonant > 0)
  for (std::vector<subgroup>::iterator s = result.subgroups.begin();
    s != result.subgroups.end(); ++s)
   s->total += nonresonant;
 return result.subgroups;
};

std::vector<subgroup> compute_subgroups(
  double total,
  const std::vector<double>& partials,
  const std::vector<subgroups::knot>& knots,
  double maxerror,
  double* error,
  int ignore
) {
 using namespace subgroups;
 
#ifdef DEBUG
 debug << "compute_subgroups:\ntotal0 = " << total
  << "\npartials = " << partials << '\n';
 for (std::vector<knot>::const_iterator k = knots.begin(); k != knots.end(); ++k)
  debug << k->dilution << '\t' << k->total1 << '\t' << k->total2 << '\t' 
   << k->partial << '\n';
 debug << '\n';
#endif

 double total0 = 0;
 for (std::vector<double>::const_iterator x = partials.begin(); 
   x != partials.end(); ++x)
  total0 += *x;

 subgroups_error result(knots, maxerror, ignore);
 result.subgroups.resize(1);
 result.subgroups[0].probability = 1;
 result.subgroups[0].total = total0;
 result.subgroups[0].partial = partials;
 result.test();

#ifdef DEBUG
 debug << "compute_subgroups: result.size() = " << result.subgroups.size()
  << ", result_error = " << result << '\n';
#endif

 if (result.enough()) {
  if (error) *error = result.error();
  result.subgroups[0].total += total - total0;
  return result.subgroups;
 };

 try {
  subgroups_error tmp(knots, maxerror, ignore);
  tmp.subgroups = compute_subgroups(
   pade(
    knots.front().dilution,
    -total0 - knots.front().dilution, 
    1, 
    knots.front().total1, 
    -knots.front().total2
   ),
   partials, 
   knots
  );
  tmp.test();
#ifdef DEBUG
  debug << "compute_subgroups: " << tmp << '\n';
  debug << tmp.subgroups;
#endif
  if (tmp.total_error < maxerror && !tmp.positive)
   tmp.minimize(knots);
  if (tmp.positive && (tmp.total_error < result.total_error || tmp.enough())) {
   result = tmp;
#ifdef DEBUG
   debug << "compute_subgroups: result.size() = " << result.subgroups.size() 
    << ", result_error = " << result << '\n';
#endif
   if (result.enough()) {
    if (error) *error = result.error();
    result.subgroups[0].total += total-total0;
    result.subgroups[1].total += total-total0;
    return result.subgroups;
   };
  }
#ifdef DEBUG
  else
   debug << "compute_subgroups: tmp.size() = " << tmp.subgroups.size() 
    << ", tmp_error = " << tmp << '\n';
#endif
 } catch (std::exception& e) {
#ifdef DEBUG
  debug << e.what() << '\n';
#endif
 };

 subgroups_error current = result;

 for (size_t K = 3; 
      K < knots.size() && !(result.enough() && result.positive); 
      ++K) {
  test_volume(total0, partials, knots, K, current);
  if (current.subgroups.empty()) continue;
//  if (current.total_error < maxerror && !current.positive)
  if (!(current.enough() && current.positive))
   current.minimize(knots);
  if (current.positive && current.error() < result.error()
      // This is a rather ugly hack to remove degenerated solutions. It would be
      // better to redesign the model
      && !current.degenerated()) {
   result = current;
#ifdef DEBUG
   debug << "compute_subgroups: result.size() = " << result.subgroups.size() 
    << ", result_error = " << result << '\n';
#endif
  } 
#ifdef DEBUG
  else
   debug << "compute_subgroups: current.size() = " << current.subgroups.size()
    << ", current_error = " << current << '\n';
#endif
 };

 if (result.enough() && result.positive) {
  double nonresonant = total - total0;
  if (nonresonant > 0)
   for (std::vector<subgroup>::iterator s = result.subgroups.begin(); 
     s != result.subgroups.end(); ++s)
    s->total += nonresonant;
  if (error) *error = result.error();
  return result.subgroups;
 };

 throw bad_accuracy(result.error(), result.subgroups);
};
